from numpy import asarray, isinf, isnan, spacing as eps, log, exp, cumsum
from numpy.random import rand
from arspy.probability_utils import exp_normalize


class HullNode(object):
    def __init__(self, m, b, left, right, pr=0.0):
        self.m, self.b = m, b
        self.left, self.right = left, right
        self.pr = pr

    def __eq__(self, other):
        from math import isclose

        def close(a, b):
            if a is b is None:
                return True
            if (a is None and b is not None) or (b is None and a is not None):
                return False
            return isclose(a, b, abs_tol=1e-02)

        return all((
            close(self.m, other.m), close(self.left, other.left),
            close(self.right, other.right),
            close(self.pr, other.pr)
        ))

    def __repr__(self):
        return "HullNode({m}, {b}, {left}, {right}, {pr})".format(
            m=self.m, b=self.b, left=self.left, right=self.right, pr=self.pr
        )

    def __hash__(self):
        return hash(str(self))


def compute_hulls(S, fS, domain):

    assert(len(S) == len(fS))
    assert(len(domain) == 2)

    lower_hull = []

    for li in range(len(S) - 1):
        m = (fS[li + 1] - fS[li]) / (S[li + 1] - S[li])
        b = fS[li] - m * S[li]
        left = S[li]
        right = S[li + 1]
        lower_hull.append(HullNode(m=m, b=b, left=left, right=right))

    # compute upper piecewise-linear hull

    # expected final length of upper hull after full computation
    n_upper_segments = 2 * (len(S) - 2) + isinf(domain[0]) + isinf(domain[1])

    upper_hull = []

    if isinf(domain[0]):
        # first line (from -infinity)

        m = (fS[1] - fS[0]) / (S[1] - S[0])
        b = fS[0] - m * S[0]
        pr = compute_segment_log_prob(float("-inf"), S[0], m, b)

        upper_hull.append(
            HullNode(m=m, b=b, pr=pr, left=float("-inf"), right=S[0])
        )

    # second line
    m = (fS[2] - fS[1]) / (S[2] - S[1])
    b = fS[1] - m * S[1]
    pr = compute_segment_log_prob(S[0], S[1], m, b)

    upper_hull.append(HullNode(m=m, b=b, pr=pr, left=S[0], right=S[1]))

    # interior lines
    # there are two lines between each abscissa
    for li in range(1, len(S) - 2):

        m1 = (fS[li] - fS[li - 1]) / (S[li] - S[li - 1])
        b1 = fS[li] - m1 * S[li]

        m2 = (fS[li + 2] - fS[li + 1]) / (S[li + 2] - S[li + 1])
        b2 = fS[li + 1] - m2 * S[li + 1]

        if isinf(m1) and isinf(m2):
            raise ValueError("both hull slopes are infinite")

        dx1 = S[li] - S[li - 1]
        df1 = fS[li] - fS[li - 1]
        dx2 = S[li + 2] - S[li + 1]
        df2 = fS[li + 2] - fS[li + 1]

        f1 = fS[li]
        f2 = fS[li + 1]
        x1 = S[li]
        x2 = S[li + 1]

        # more numerically stable than above
        ix = ((f1 * dx1 - df1 * x1) * dx2 - (f2 * dx2 - df2 * x2) * dx1) / (df2 * dx1 - df1 * dx2)

        if isinf(m1) or abs(m1 - m2) < 10.0 ** 8 * eps(m1):
            ix = S[li]
            pr1 = float("-inf")
            pr2 = compute_segment_log_prob(ix, S[li + 1], m2, b2)
        elif isinf(m2):
            ix = S[li + 1]
            pr1 = compute_segment_log_prob(S[li], ix, m1, b1)
            pr2 = float("-inf")
        else:
            if isinf(ix):
                raise ValueError("Non finite intersection")

            if abs(ix - S[li]) < 10.0 ** 12 * eps(S[li]):
                ix = S[li]
            elif abs(ix - S[li + 1]) < 10.0**12 * eps(S[li + 1]):
                ix = S[li + 1]

            if ix < S[li] or ix > S[li + 1]:
                raise ValueError("Intersection out of bounds -- logpdf is not concave")

            pr1 = compute_segment_log_prob(S[li], ix, m1, b1)
            pr2 = compute_segment_log_prob(ix, S[li + 1], m2, b2)

            upper_hull.append(HullNode(m=m1, b=b1, pr=pr1, left=S[li], right=ix))
            upper_hull.append(HullNode(m=m2, b=b2, pr=pr2, left=ix, right=S[li + 1]))

    # second last line
    m = (fS[-2] - fS[-3]) / float(S[-2] - S[-3])
    b = fS[-2] - m * S[-2]
    pr = compute_segment_log_prob(S[-2], S[-1], m, b)

    upper_hull.append(HullNode(m=m, b=b, pr=pr, left=S[-2], right=S[-1]))

    if isinf(domain[1]):
        # last line (to infinity)
        m = (fS[-1] - fS[-2]) / (S[-1] - S[-2])
        b = fS[-1] - m * S[-1]
        pr = compute_segment_log_prob(S[-1], float("inf"), m, b)

        upper_hull.append(
            HullNode(m=m, b=b, pr=pr, left=S[-1], right=float("inf"))
        )

    # normalize probabilities
    normalized_probabilities = exp_normalize(
        asarray([node.pr for node in upper_hull])
    )

    for node, probability in zip(upper_hull, normalized_probabilities):
        node.pr = probability

    assert(len(lower_hull) == len(S) - 1)
    assert(len(upper_hull) == n_upper_segments)

    return lower_hull, upper_hull


def compute_segment_log_prob(l, r, m, b):
    if l == float("-inf"):
        return -log(m) + m * r + b
    elif r == float("inf"):
        return -log(-m) + m * l + b

    M = max(m * r + b, m * l + b)

    return -log(abs(m)) + log(abs(exp(m * r + b - M) - exp(m * l + b - M))) + M


def sample_upper_hull(upper_hull):
    cdf = cumsum([node.pr for node in upper_hull])

    # randomly choose a line segment
    U = rand()

    node = next(
        (node for node, cdf_value in zip(upper_hull, cdf) if U < cdf_value),
        upper_hull[-1]  # default is last line segment
    )

    # sample along that line segment
    U = rand()

    m, left, right = node.m, node.left, node.right

    M = max(m * right, m * left)
    x = (log(U * (exp(m * right - M) - exp(m * left - M)) + exp(m * left - M)) + M) / m

    assert(x >= left and x <= right)

    if isinf(x) or isnan(x):
        raise ValueError("sampled an infinite or 'nan' x")

    return x


def evaluate_hulls(x, lower_hull, upper_hull):

    # lower bound
    lh_val = 0.0

    if x < min(node.left for node in lower_hull):
        lh_val = float("-inf")
    elif x > max(node.right for node in lower_hull):
        lh_val = float("-inf")
    else:
        node = next(node for node in lower_hull if node.left <= x <= node.right)
        lh_val = node.m * x + node.b

    # upper bound
    uh_val = 0.0

    node = next(node for node in upper_hull if node.left <= x <= node.right)
    uh_val = node.m * x + node.b

    return lh_val, uh_val



from numpy import log, exp, allclose, sqrt
from numpy.random import seed
import numpy as np

from os.path import dirname, realpath, join

from arspy.ars import adaptive_rejection_sampling


data_file = "{}/ars_{{}}.npy".format(
    join(dirname(realpath(__file__)), "reference_data", "ars_data")
).format


def gaussian(x, sigma=1):
    return log(exp(-x ** 2 / sigma))


def half_gaussian(x, sigma=3):
    return log(exp(-x ** 2 / sigma)) * (1 * (x <= 0) + 1e300 * (x > 0))


def relativistic_momentum_logpdf(p, m=1., c=1.):
    return -m * c ** 2 * sqrt(p ** 2 / (m ** 2 * c ** 2) + 1)


tests = {
    "1d-gaussian": {"name": "1d-gaussian",
                    "data": data_file("gaussian"),
                    "func": gaussian,
                    "a": -2, "b": 2,
                    "domain": (float("-inf"), float("inf")),
                    "n_samples": 20},
    "1d-half-gaussian": {"name": "1d-half-gaussian",
                         "data": data_file("half_gaussian"),
                         "func": half_gaussian,
                         "a": -2, "b": 0,
                         "domain": [float("-inf"), 0],
                         "n_samples": 20},
    "relativistic_monte_carlo_logpdf": {
        "name": "relativistic_momentum_logpdf",
        "data": data_file("relativistic_logpdf"),
        "func": relativistic_momentum_logpdf,
        "a": -10.0, "b": 10.0,
        "domain": [float("-inf"), float("inf")],
        "n_samples": 20
    }

}


def _run(test_name):
    seed(1)
    input_dict = tests[test_name]

    # name = input_dict["name"]
    a = input_dict["a"]
    b = input_dict["b"]
    domain = input_dict["domain"]
    n_samples = input_dict["n_samples"]

    logpdf = input_dict["func"]

    python_result = adaptive_rejection_sampling(
        logpdf=logpdf, a=a, b=b, domain=domain, n_samples=n_samples
    )

    # load old result computed by other implementation (julia)
    julia_result = np.load(input_dict["data"])

    assert(allclose(julia_result, python_result, atol=3e-01))


def test_gaussian():
    _run("1d-gaussian")


def test_half_gaussian():
    _run("1d-half-gaussian")


def test_relativistic_monte_carlo_logpdf():
    _run("relativistic_monte_carlo_logpdf")



from numpy import sign, log, unique, linspace, isinf
from numpy.random import rand
from arspy.hull import compute_hulls, evaluate_hulls, sample_upper_hull


def adaptive_rejection_sampling(logpdf,
                                a: float, b: float,
                                domain,
                                n_samples: int):

    assert(hasattr(logpdf, "__call__"))
    assert(len(domain) == 2), "Domain must be two-element iterable."
    assert(domain[1] >= domain[0]), "Invalid domain, it must hold: domain[1] >= domain[0]."
    assert(n_samples >= 0), "Number of samples must be >= 0."

    if a >= b or isinf(a) or isinf(b) or a < domain[0] or b > domain[1]:
        raise ValueError("invalid a and b")

    n_derivative_steps = 1e-3 * (b - a)

    S = (a, a + n_derivative_steps, b - n_derivative_steps, b)

    if domain[0] == float("-inf"):
        # ensure positive derivative at 'a'
        derivative_sign = sign(logpdf(a + n_derivative_steps) - logpdf(a))
        positive_derivative = derivative_sign > 0

        assert(positive_derivative), "derivative at 'a' must be positive, since the domain is unbounded to the left"

    if domain[1] == float("inf"):
        # ensure negative derivative at 'b'
        derivative_sign = sign(logpdf(b) - logpdf(b - n_derivative_steps))
        negative_derivative = derivative_sign < 0

        assert(negative_derivative), "derivative at 'b' must be negative, since the domain is unbounded to the right"

    # initialize a mesh on which to create upper & lower hulls
    n_initial_mesh_points = 3

    S = unique(
        (S[0], *(linspace(S[1], S[2], num=n_initial_mesh_points + 2)), S[3])
    )

    fS = tuple(logpdf(s) for s in S)

    lower_hull, upper_hull = compute_hulls(S=S, fS=fS, domain=domain)

    samples = []

    while len(samples) < n_samples:

        mesh_changed = False

        x = sample_upper_hull(upper_hull)

        lh_val, uh_val = evaluate_hulls(x, lower_hull, upper_hull)

        U = rand()

        if log(U) <= lh_val - uh_val:
            # accept u is below lower bound
            samples.append(x)

        elif log(U) <= logpdf(x) - uh_val:
            # accept, u is between lower bound and f
            samples.append(x)

            mesh_changed = True
        else:
            # reject, u is between f and upper_bound
            mesh_changed = True

        if mesh_changed:
            S = sorted([*S, x])
            fS = tuple(logpdf(s) for s in S)
            lower_hull, upper_hull = compute_hulls(S=S, fS=fS, domain=domain)
    return samples



from os.path import realpath, dirname, join as path_join
from setuptools import setup, find_packages


name = "ARSpy"
description = "ARSpy"
long_description = "Adaptive Rejection Sampling for Python - Matlab Style"
maintainer = "Moritz Freidank"
maintainer_email = "freidankm@googlemail.com"

url = "https://github.com/MFreidank/ARSpy"

version = "0.2"

project_root = dirname(realpath(__file__))
requirements_file = path_join(project_root, "requirements.txt")

with open(requirements_file, "r") as f:
    install_requirements = f.read().splitlines()

setup_requirements = ["pytest-runner"]
test_requirements = ["pytest", "pytest-cov"]


if __name__ == "__main__":
    setup(
        name=name,
        version=version,
        maintainer=maintainer,
        maintainer_email=maintainer_email,
        description=description,
        url=url,
        download_url="https://github.com/MFreidank/ARSpy/archive/0.1.tar.gz",
        long_description=long_description,
        packages=find_packages(),
        keyword=["sampling", "adaptive rejection sampling", "adaptive", "rejection", "ars"],
        # package_data={"docs": ["*"]},
        # include_package_data=True,
        install_requires=install_requirements,
        setup_requires=setup_requirements,
        tests_require=test_requirements,
    )
